﻿using System;
using System.Collections.Generic;
using UnityEngine;
using MacGruber;

// ***********************************************************************************
// SallySoundFX
// ***********************************************************************************
// Created by Sally Whitemane on VAM Hub
// Licensed under CC BY-SA - https://creativecommons.org/licenses/by-sa/4.0/
// 
// Dependency:
// MacGruber.LogicBricks.12:/Custom/Scripts/MacGruber/LogicBricks/Internal/MacGruber_Utils.cs

namespace Sally
{
    public class Dancefloor : MVRScript
    {
        private Vector3[] vertices;
        private Vector2[] uv;
        private int[] triangles;
        private Color[] colors;
        private Mesh mesh;
        private Material material;
        private GameObject go;

        private bool firstStart = true;
        private int xSize = 8;
        private int zSize = 8;
        private float xPos = 0;
        private float yPos = 0;
        private float zPos = 0;
        private JSONStorableBool jbActive;
        private JSONStorableStringChooser jsColorEffectChooser;
        private JSONStorableFloat jfColorEffectAmp;
        private JSONStorableFloat jfXSize;
        private JSONStorableFloat jfZSize;
        private JSONStorableFloat jfXPos;
        private JSONStorableFloat jfYPos;
        private JSONStorableFloat jfZPos;

        private int colorEffect = 2;
        private int count = 0;
        private int countSpeed = 1;
        private float deltaTimeCounter = 0f;
        private float[] sin1x = LUT.CreateSinWave(256, 1, true);

        private void Start()
        {
            try
            {
                //SuperController.LogMessage("SallySoundFX Dancefloor Start()");
                go = new GameObject("DanceFloor", typeof(MeshFilter), typeof(MeshRenderer));
                UpdateMesh();
                go.GetComponent<MeshFilter>().mesh = mesh;

                // SHADER MUST SUPPORT VERTEX COLORS
                // Standard               -> texture,    shadow, NO vertex colors
                // Legacy Shaders/Diffuse -> texture, no shadow, NO vertex colors
                // Sprites/Default        -> texture, no shadow
                // Particles/Additive     -> texture, no shadow, looks more diffused
                // UI/Default             -> texture, no shadow
                // Unlit/Texture          -> NOT found
                // Sprites/Diffuse        -> NOT found
                Shader shader = Shader.Find("Sprites/Default");
                //SuperController.LogMessage("SallySoundFX Dancefloor Start() Shader name: " + shader.name);
                material = new Material(shader);
                if (material == null)
                {
                    SuperController.LogMessage("SallySoundFX Dancefloor Start() ERROR: material == null");
                    return;
                }
                //SuperController.LogMessage("SallySoundFX Dancefloor Start() material: " + material.name);
                material.color = Color.white;

				// LOAD TEXTURE
				//Texture2D tex = new Texture2D(128, 128);
				Texture2D tex = new Texture2D(128, 128, TextureFormat.DXT1, false);
				tex.LoadRawTextureData(TextureRaw.dancefloorTileDXT1);
				tex.Apply();
				if (tex == null)
                {
                    SuperController.LogMessage("SallySoundFX Dancefloor Start() ERROR: tex == 0");
                    return;
                }
                //SuperController.LogMessage("SallySoundFX Dancefloor Start() texture: " + tex.name);
                
                material.SetTexture("_MainTex", tex);
                go.GetComponent<MeshRenderer>().material = material;
                go.transform.position = new Vector3(xPos,yPos,zPos);
                go.transform.localScale = new Vector3(0.5f, 0.5f, 0.5f);
            }
            catch (Exception e)
            {
                Debug.Log("Exception: " + e);
            }
        }

        public override void Init() {
            try {
                if (firstStart)
                {
                    firstStart = false;
                    jbActive = Utils.SetupToggle(this, "Active", true, false);
                    jbActive.setCallbackFunction += ToggleActive;

                    JSONStorableString jsText = new JSONStorableString("Text", "The dancefloor is an experimental visualization. It uses internal frequency bands. They use bad hardcoded amplification values. So they are from from optimal. Everything is to flickery and needs to be smoother.");
                    CreateTextField(jsText, false);

                    List<string> colorEffects = new List<string>();
                    colorEffects.Add("Red Blue Waves");
                    colorEffects.Add("Red Blue Bands");
                    colorEffects.Add("Rainbow Bands");
                    colorEffects.Add("Rainbow Waves");
                    jsColorEffectChooser = Utils.SetupStringChooser(this, "Color Effects", colorEffects, false);
                    jsColorEffectChooser.setCallbackFunction += SetColorEffect;
                    jfColorEffectAmp = Utils.SetupSliderFloat(this, "Color Amplification", 1f, 0.1f, 20f, false);

                    jfXSize = Utils.SetupSliderInt(this, "x Size", 8, 1, 16, true);
                    jfXSize.setCallbackFunction += SetXSize;
                    jfZSize = Utils.SetupSliderInt(this, "z Size", 8, 1, 16, true);
                    jfZSize.setCallbackFunction += SetZSize;

                    xPos = -xSize * 0.25f;
                    yPos = 0f;
                    zPos = -zSize * 0.25f;
                    jfXPos = Utils.SetupSliderFloat(this, "x Position", xPos, -5f, 5f, true);
                    jfXPos.setCallbackFunction += SetXPos;
                    jfYPos = Utils.SetupSliderFloat(this, "y Position", yPos, -5f, 5f, true);
                    jfYPos.setCallbackFunction += SetYPos;
                    jfZPos = Utils.SetupSliderFloat(this, "z Position", zPos, -5f, 5f, true);
                    jfZPos.setCallbackFunction += SetZPos;
                }
            }
			catch (Exception e) {
				SuperController.LogError("Exception: " + e);
			}
        }

        private void Update()
        {
            List<Color> colors = new List<Color>();
            int tile = 0;
            int band = 0;
            float amp = jfColorEffectAmp.val;
            // limit count increase to to 30 times per frames (or else it's to fast)
            deltaTimeCounter += Time.deltaTime;
            if (deltaTimeCounter > 1f / 30f)
            {
                deltaTimeCounter = deltaTimeCounter - 1f / 30f;
                count += countSpeed;
                //SuperController.LogMessage(deltaTimeCounter.ToString("0.0000")+ " " + count);
            }
            switch (colorEffect)
            {
                default:
                case 0: // [TESTED] Red Blue Waves
                    if (count >= sin1x.Length)
                        count = 0;
                    for (int v = 0; v < vertices.Length; v += 4)
                    {
                        if (count + tile >= sin1x.Length)
                            tile -= sin1x.Length;
                        Color c = new Color(0.05f + 0.95f * sin1x[count + tile] * SoundFX.freqBands[band] * amp, 0.0f, 0.05f + 0.95f * SoundFX.freqBands[band] * amp);
                        colors.Add(c);
                        colors.Add(c);
                        colors.Add(c);
                        colors.Add(c);
                        tile++;
                        band++;
                        if (band >= SoundFX.freqBands.Length)
                            band = 0;
                    }
                    break;
                case 1: // [TESTED] Red Blue Bands
                    for (int v = 0; v < vertices.Length; v += 4)
                    {
                        Color c = new Color(0.05f + 2f * SoundFX.freqBands[band] * amp, 0, amp * 1f - SoundFX.freqBands[band] * amp);
                        colors.Add(c);
                        colors.Add(c);
                        colors.Add(c);
                        colors.Add(c);
                        tile++;
                        band++;
                        if (tile > 16)
                            tile = 0;
                        if (band >= SoundFX.freqBands.Length)
                            band = 0;
                    }
                    break;
                case 2: // [TESTED] Rainbow Bands
                    for (int v = 0; v < vertices.Length; v += 4)
                    {
                        
                        Color c = Color.HSVToRGB( (band+1) * 0.125f, 1f, 0.1f + 0.9f*SoundFX.freqBands[band] * jfColorEffectAmp.val);
                        colors.Add(c);
                        colors.Add(c);
                        colors.Add(c);
                        colors.Add(c);
                        band++;
                        if (band >= SoundFX.freqBands.Length)
                            band = 0;
                    }
                    break;
                case 3: // [TESTED] Rainbow Waves
                    if (count >= sin1x.Length)
                        count = 0;
                    for (int v = 0; v < vertices.Length; v += 4)
                    {
                        if (count + tile >= sin1x.Length)
                            tile -= sin1x.Length;
                        Color c = Color.HSVToRGB(sin1x[count + tile], 1f, 0.5f + 0.5f * SoundFX.freqBands[band] * amp);
                        colors.Add(c);
                        colors.Add(c);
                        colors.Add(c);
                        colors.Add(c);
                        tile++;
                        band++;
                        if (band >= SoundFX.freqBands.Length)
                            band = 0;
                    }
                    break;
            }
            mesh.SetColors(colors);
        }

        void OnDestroy()
        {
            Destroy(go);
        }

        void UpdateMesh()
        {
            Debug.Log("SallySoundFX Dancefloor UpdateMesh()");
            //mesh.Clear();

            vertices = new Vector3[xSize * zSize * 4];
            uv = new Vector2[xSize * zSize * 4];
            triangles = new int[xSize * zSize * 6];
            colors = new Color[vertices.Length];

            Debug.Log("DEBUG: vertices.Length: " + vertices.Length);
            Debug.Log("DEBUG: uv.Length: " + uv.Length);
            Debug.Log("DEBUG: triangles.Length: " + triangles.Length);

            int vCount = 0;
            int vIndex = 0;
            for (int z = 0; z < zSize; z++)
                for (int x = 0; x < xSize; x++)
                {
                    vertices[vCount] = new Vector3(x, 0f, z + 1f);
                    vertices[vCount + 1] = new Vector3(x + 1f, 0f, z + 1f);
                    vertices[vCount + 2] = new Vector3(x, 0f, z + 0f);
                    vertices[vCount + 3] = new Vector3(x + 1f, 0f, z + 0f);

                    uv[vCount] = new Vector2(0f, 1f);
                    uv[vCount + 1] = new Vector2(1f, 1f);
                    uv[vCount + 2] = new Vector2(0f, 0f);
                    uv[vCount + 3] = new Vector2(1f, 0f);

                    triangles[vIndex] = vCount;
                    triangles[vIndex + 1] = vCount + 1;
                    triangles[vIndex + 2] = vCount + 2;
                    triangles[vIndex + 3] = vCount + 2;
                    triangles[vIndex + 4] = vCount + 1;
                    triangles[vIndex + 5] = vCount + 3;
                    vCount += 4;
                    vIndex += 6;
                }
            mesh = new Mesh();
            mesh.vertices = vertices;
            mesh.uv = uv;
            mesh.triangles = triangles;
            mesh.colors = colors;
            mesh.name = "DanceFloorMesh";
        }

        //////////////////////////////////////////////////////////////////////////////////
        // GUI Callbacks
        //////////////////////////////////////////////////////////////////////////////////
        private void ToggleActive(bool b) {
            if (b)
                Activate();
            else
                Deactivate();
        }

        private void SetColorEffect(string s) {
            switch (s)
            {
                default:
                case "Red Blue Waves":
                    colorEffect = 0;
                    countSpeed = 4;
                    break;
                case "Red Blue Bands":
                    colorEffect = 1;
                    countSpeed = 1;
                    break;
                case "Rainbow Bands":
                    colorEffect = 2;
                    countSpeed = 1;
                    break;
                case "Rainbow Waves":
                    colorEffect = 3;
                    countSpeed = 4;
                    break;
            }
        }

        private void SetXSize(float v) {
            xSize = (int)v;
            xPos = -xSize * 0.25f;
            // accessing jfXPos causes this exception during scene load even though the object seems valid ?!
            // Exception during RestoreFromLast of plugin#0_Sally.Dancefloor: System.NullReferenceException
            //SuperController.LogError("jfXPos " + jfXPos.ToString());
            //SuperController.LogError("jfXPos.val " + jfXPos.val.ToString());
            //SuperController.LogError("jfXPos.locked " + jfXPos.locked.ToString());
            //SuperController.LogError("xPos " + xPos.ToString());
            try
            {
                jfXPos.val = xPos;
                jfXPos.defaultVal = xPos;
            }
            catch (NullReferenceException) { }
            UpdateMesh();
            if (go != null)
            {
                go.GetComponent<MeshFilter>().mesh = mesh;
                go.transform.position = new Vector3(xPos, 0f, zPos);
            }
        }

        private void SetZSize(float v) {
            zSize = (int)v;
            zPos = -zSize * 0.25f;
            try
            {
                jfZPos.val = zPos;
                jfZPos.defaultVal = zPos;
            }
            catch (NullReferenceException) { }
            UpdateMesh();
            if (go != null)
            {
                go.GetComponent<MeshFilter>().mesh = mesh;
                go.transform.position = new Vector3(xPos, 0f, zPos);
            }
        }
        private void SetXPos(float v) {
            xPos = v;
            jfXPos.defaultVal = -xSize * 0.25f;
            go.transform.position = new Vector3(xPos, yPos, zPos);
        }
        private void SetYPos(float v) {
            yPos = v;
            go.transform.position = new Vector3(xPos, yPos, zPos);
        }
        private void SetZPos(float v) {
            zPos = v;
            jfZPos.defaultVal = -zSize * 0.25f;
            go.transform.position = new Vector3(xPos, yPos, zPos);
        }

        //////////////////////////////////////////////////////////////////////////////////
        // Helper functions
        //////////////////////////////////////////////////////////////////////////////////
        private void Activate() {
            Start();
        }
        private void Deactivate() {
            OnDestroy();
        }
    }
}